#include <stdint.h>
#include "../pdl.h"

typedef int32_t fixed_t;
#define NB_POINT_CUBE 8
#define NB_LINES_CUBE 12

#define PRECISION 8
#define TO_FIXED(num) (int)((num)*(1<<PRECISION))
#define TO_INT(num) ((num)>>PRECISION)
#define MUL(a, b) ((a*b)>>PRECISION)
#define DIV(a, b) ((a<<PRECISION)/b)
#define SGN(a) ((a) < 0 ? -1 : (a) > 0 ? 1 : 0)
#define C_RGB(r,g,b) (((r) << 11) + ((g) << 6) + (b))

uint8_t Screen256[ 88704 ];
uint16_t Palette256[ 256 ];

fixed_t angle = 0;

typedef struct
{
  fixed_t x;
  fixed_t y;
  fixed_t z;
  int16_t sx;
  int16_t sy;
} Vertex3D;

typedef struct
{
    int8_t p1;
    int8_t p2;
} Pairs;

Pairs Lines[NB_LINES_CUBE] = {
    { 0, 1 },
    { 1, 2 },
    { 2, 3 },
    { 3, 0 },

    { 4, 5 },
    { 5, 6 },
    { 6, 7 },
    { 7, 4 },

    { 0, 4 },
    { 1, 5 },
    { 2, 6 },
    { 3, 7 },
};
Vertex3D Cube[NB_POINT_CUBE];


fixed_t sine(fixed_t d) {
    fixed_t v;
    fixed_t org_d = d % TO_FIXED(360);
    while(org_d < TO_FIXED(0)) org_d += TO_FIXED(360);
    d = org_d % TO_FIXED(180);
    v = DIV(MUL(MUL(TO_FIXED(4), d), (TO_FIXED(180)-d)),
            (TO_FIXED(40500)-MUL(d, (TO_FIXED(180)-d))));
    return (org_d > TO_FIXED(180) ? v : -v);
}

fixed_t cosine(fixed_t d) {
    return sine( d + TO_FIXED(90) );
}

void RotateAndProjectVertex( Vertex3D *vertex, fixed_t angle )
{
    // Ruse du cowboy car dans notre cas ax=0 et ay=az=angle;
	fixed_t ry = vertex->y;
	fixed_t rz = vertex->z;
    fixed_t C = cosine(angle), S = sine(angle);
	fixed_t rx = MUL(vertex->x, C) + MUL(rz, S);
	rz =  MUL(rz, C) - MUL(vertex->x, S);
    fixed_t tmpx = MUL(rx, C) - MUL(ry, S);
	ry = MUL(rx, S) + MUL(ry, C);
	rx = tmpx;

    fixed_t divisor = DIV( TO_FIXED(250), (rz +  TO_FIXED(4.16667)) );
    vertex->sx = TO_INT( MUL( rx, divisor ) );
    vertex->sy = TO_INT( MUL( ry, divisor ) );
}

void Drawline( uint8_t *buffer, int x1, int y1, int x2, int y2, uint8_t color)
{
	int i;
	int dx = x2 - x1;
    int dy = y2 - y1;
	int sx = SGN(dx);
    int sy = SGN(dy);

    buffer += 396 * (y1 + 112) + (x1 + 198);

	dx = (dx >= 0 ? dx : -dx);
    dy = (dy >= 0 ? dy : -dy);

    if(dx < dy) {
        int tmp = dx;
        dx = dy;
        dy = tmp;
        tmp = sx;
        sx = 396 * sy;
        sy = tmp;
    }
    else {
        sy *= 396;
    }

	int cumul = dx >> 1;

	for(i = 0; i <= dx; i++)
	{
		buffer += sx;
		cumul += dy;
		if(cumul > dx) cumul -= dx, buffer += sy;
        *buffer = color;
	}
}

void Blur( uint8_t *buffer )
{
	uint16_t temp;
    buffer += 88308;
    
    for( int index = 0; index < 223; index++ )
    {
        for( int x = 0;  x < 394; x++ )
        {
            temp = buffer[0] +
                   buffer[1] +
                   buffer[2] +
                   buffer[-395];
            buffer[-395] = temp >> 2;
            buffer++;
        }
        buffer -= 395 + 395;
    }
}

void Dclear( uint8_t *Screen, uint8_t color )
{
	for(int i = 0; i < 88704; i++)
		Screen[i] = color;
}

void Render( uint16_t *vram, uint8_t *Screen, uint16_t *Palette )
{
	for(int i = 0; i < 88704; i++)
		vram[i] = Palette[ Screen[i] ];
}

void SetPalette( uint16_t *Palette )
{
	for(int u = 0; u<64; u++)
    {
        Palette[u]     = C_RGB(   0,   0,   0 );
        Palette[u+64]  = C_RGB( u/2,   0,   0 );
        Palette[u+128] = C_RGB(  31, u/2,   0 );
        Palette[u+192] = C_RGB(  31,  31, u/2 );
    }
}

void FirecubeInit( void )
{
    /* Initialize longwords of Cube (4 per vertex) to either -0.75 or 0.75,
       as indicated by one bit of initMask. We initialize the pair sx/sy using
       a useless bit. sx/sy is normally uninitialized, so we don't care.

         Order:   sx/sy  z  y  x
         Cube[0]:     0  -  -  -
         Cube[1]:     0  -  -  +
         Cube[2]:     0  -  +  +
         Cube[3]:     0  -  +  -
         Cube[4]:     0  +  -  -
         Cube[5]:     0  +  -  +
         Cube[6]:     0  +  +  +
         Cube[7]:     0  +  +  - */
    int initMask = 0b01110110010001010011001000000001;
    int x0 = TO_FIXED(0.75), x1 = -x0;
    int *ptr = (int *)Cube;

    do {
        /* This could use shlr beautifully */
        *ptr++ = (initMask & 1) ? x1 : x0;
        initMask >>= 1;
    }
    while(initMask);

/*
    Cube[0].x = TO_FIXED(-0.75);    Cube[0].y = TO_FIXED(-0.75);    Cube[0].z = TO_FIXED(-0.75); 
    Cube[1].x = TO_FIXED(0.75);     Cube[1].y = TO_FIXED(-0.75);    Cube[1].z = TO_FIXED(-0.75); 
    Cube[2].x = TO_FIXED(0.75);     Cube[2].y = TO_FIXED(0.75);     Cube[2].z = TO_FIXED(-0.75); 
    Cube[3].x = TO_FIXED(-0.75);    Cube[3].y = TO_FIXED(0.75);     Cube[3].z = TO_FIXED(-0.75); 

    Cube[4].x = TO_FIXED(-0.75);    Cube[4].y = TO_FIXED(-0.75);    Cube[4].z = TO_FIXED(0.75); 
    Cube[5].x = TO_FIXED(0.75);     Cube[5].y = TO_FIXED(-0.75);    Cube[5].z = TO_FIXED(0.75); 
    Cube[6].x = TO_FIXED(0.75);     Cube[6].y = TO_FIXED(0.75);     Cube[6].z = TO_FIXED(0.75); 
    Cube[7].x = TO_FIXED(-0.75);    Cube[7].y = TO_FIXED(0.75);     Cube[7].z = TO_FIXED(0.75); 
*/
}

void FirecubeUpdate( void )
{
	angle += 1;
    if(angle == 360)
        angle = 0;
    for (int i = 0; i < NB_POINT_CUBE; i++)
    {
		RotateAndProjectVertex(&Cube[i], TO_FIXED(angle));
    }
}

void FirecubeRender( uint8_t *screen )
{ 
    for (int j = 0; j < NB_LINES_CUBE; j++)
        Drawline( screen, Cube[ Lines[j].p1 ].sx, Cube[ Lines[j].p1 ].sy, Cube[ Lines[j].p2 ].sx, Cube[ Lines[j].p2 ].sy, 255 );
    Blur( screen );
}

PDL_MAIN int main(void)
{
	SetPalette( Palette256 );
	Dclear( Screen256, 0 );

	FirecubeInit( );

    do {
		FirecubeUpdate( );
		FirecubeRender( Screen256 );
		Render( PDL->vram, Screen256, Palette256 );
		PDL->dupdate();
	}
	while(PDL->time_ms <= 10000);

	return 0;
}
